%==========================================================================
%==========================================================================
% Copyright (C) 2014 by Arne Leijon
% Email: leijon@kth.se

% This program is free software for academic and non-commercial
% use, to be used at user's own risk, no guarantee of performance.
% The user is permitted to copy, distribute and transmit only unaltered
% copies of this program. 
%==========================================================================
%==========================================================================
%GMMB - Bayesian version of Gaussian Mixture Model,
%i.e. a Probability distribution class, representing
%a weighted mixture of Gaussian probability distributions,
%each defined by a GaussB object.
%
%A Gaussian mixture can be used to approximate any probability distribution.
%
%A GMMB object can represent random vectors with any form of 
%statistical dependence between vector elements, even if each of its
%GaussB components does not allow correlated vector elements.
%
%Several GMMB objects may be collected in a multidimensional array,
%           (even if they do not have the same DataSize).
%Cell arrays are not (yet!) allowed.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%***** Usage Example:
%
%gm1= init(GMMB(3),xTraining1);%initialize a 3-component GMMB object
%gm1= adapt(gm1, xTraining1);% iterative adaptation to first data set
%gm2= init(GMMB(2),xTraining2a);%initialize another GMMB object
%gm2= adapt(gm2, xTraining2a);% iterative adaptation
%gm2= adapt(gm2, xTraining2b);% further adaptation to more data
%lP= logprob([gm1,gm2], xTest);% posterior probabilities for test vector(s)
%
%These logprob values can then be used for classification.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%Arne Leijon 2009-05-06 not fully implemented

classdef GMMB
    properties (Dependent=true)
        DataSize;
    end
    properties
        Gaussians=GaussB;%array of GaussB components (default=just 1 comp)
        MixWeight=DiscreteB(1);%mixture weight distribution
        UserData=[];%anything
    end
    methods
        function pD = GMMB(varargin)%constructor
            args=varargin;
            while ~isempty(args)
                a=args{1};
                switch class(a)
                    case 'GMMB'
                        pD= a;%just Copy it
                    case 'GaussB'
                        pD.Gaussians=a;%store it
                    case 'DiscreteB'
                        pD.MixWeight=a;%store it
                    case 'double'
                        if numel(a)==1%assume it is desired number of components
                            pD.Gaussians=repmat(GaussB,a,1);%clone trivial GaussB to right length
                            pD.MixWeight = DiscreteB(ones(length(pD.Gaussians),1));%equal weights
                        end;
                    otherwise%do nothing
                        warning('Wrong type of argument');
                end
                args=args(2:end);
            end
        end
        function s=get.DataSize(pD)
            s=length(pD.Gaussians(1).Mean);
        end
        [X,iZ]=rand(pD,nSamples);%generate random vector samples
        [f,logS]=prob(pD,x);%give probability densities for observed vector samples
        logP=logprob(pD,x);%give log(probability densities) for observed vector samples
%*****  Display methods
        plotH=plotCross(pD,proj,colors);%plot indicators of each mixed GaussB sub-object
        plotH=plotProbDensity(pD,proj,axLimit,isoProb,colors);%3D plot of probability density
%*****  Training Methods:
        pD=init(pD,x);%initialize crudely to allow future training with given data.
        r=initMixProb(pD,x);%initial partition of training data
        [pD,LLsave,r]=adapt(pDprior,x, nIter, minLLstep, callFcn);%train object properties       
    end
    methods (Access='private')
        OK=propertiesOK(gD);%check properties for consistency
    end
end
