%[pD,LLsave]=adapt(pDprior,x, nIter, minLLstep, callFcn)
%method to adapt a single GMMB object
%using observed training data.
%
%Usage:
%A GMMB object can be trained sequentially,
%using several different training data sets, as
%pD=init(GMMB(nComponents), x1);
%pD=adapt(pD,x1);
%pD=adapt(pD,x2);
%etc.....
%We might also want to allow a forgetting factor....???
%
%Input:
%pDprior=   single GMMB object, with PRIOR property values
%x=         matrix with training (column) vectors
%nIter=     (optional) min number of iterations
%minLLstep= (optional) min log-likelihood improvement per training data vector
%callFcn=   (optional) function handle to be called after each iteration,
%           for general user purpose, e.g. to plot intermediate results.
%           The function handle is always called with a single argument (pD).
%           It can be a handle to an anonymous function, e.g.
%           @(g) plotdata(g, x, arg1,etc...), allowing additional
%           externally defined arguments.
%
%Result:
%pD=        adapted version of the GMMB object
%LLsave=    vector with log-likelihood values,
%           must increase during adaptation.
%******** LL bound calculation not yet implemented in sub-classes *****
%******** so we must use a fixed number of iterations
%
%Theory and Method:
%We apply variational approximation (Bishop, sec. 10.2).
%The sub-object GaussB array pD.Gaussians, and
%the sub-object DiscreteB pD.MixWeight, have their own adapt methods.
%We just call these alternatingly, until convergence.
%However, if the prior was not trained before, we must first
%separate the training data in some way....
%
%References:
%   Bishop (2006) Pattern Recognition and Machine learning, chap 10.2.
%
%Arne Leijon 2011-05-05 slightly tested

function [pD,LLsave,r]=adapt(pDprior,x, nIter, minLLstep, callFcn)
if numel(pDprior)>1
    error('Method works only for a single object');
end;
if nargin<3
    nIter=30;%*** good default????
end;
if nargin<4||isempty(minLLstep)
    minLLstep=Inf;%to prevent any more than nIter training rounds
end;
nIter=max(nIter,1);
pD=pDprior;
if notPreTrained(pD)%this is first training data set
    [r]=initMixProb(pD,x);%initial partition of training data
    lPz=zeros(size(r,1),1);
    [pD.MixWeight,LLmix,pDWinit]=adapt(pDprior.MixWeight,r);
    pDWeight=pD.MixWeight;
%     E2k=zeros(size(r,1),1);
    pD.Gaussians=adapt(pDprior.Gaussians,x,r,lPz,pDWeight,pDWinit);
%     pD=adapt(pDprior,x,r);
    
end;
LL=-Inf;%low bound on log-likelihood of training data
LLsave=[];
while nIter>0 || LL > LLold+minLLstep
    %use better convergence criterion...???
    [r,lPz,E2k]=calcMixProb(pD,x);%normalized component responsibilities
    if any(isnan(r(:)))
        error('TEST: r= NaN');%******************
    end;
    [pD.MixWeight,LLmix,pDWinit]=adapt(pDprior.MixWeight,r);
    %mix=pD.MixWeight%********TEST print
     pDWeight=pD.MixWeight;
    [pD.Gaussians,LLgauss]=adapt(pDprior.Gaussians,x,r,lPz,pDWeight,pDWinit,E2k);
    LLold=LL;%remember previous value
    LL=LLmix+LLgauss;%new low bound on Log-Likelihood
    LLsave=[LLsave,LL];%accumulate log-likelihood
    nIter=nIter-1;
    %%%%%%%%%%%%%%%
    if nargin>=5 && ~isempty(callFcn)
        callFcn(pD);%callback function for any purpose
    end;
end;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function L=notPreTrained(pD)
%if sum of prior Weight is small,
%we have not applied any training data before:
nObs=sum(pD.MixWeight.Weight);%sum of # previous data
L=nObs<1;%default initial beta should be between 0 and 1.
%*** Maybe there is a smarter way to do this...???

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [r, lPz, E2k]=calcMixProb(pD,x)
%r(i,n)=P[Z(n)==i | x(:,n)]
%where Z(n) is the index of the sub-Gaussian that
%might have generated the vector x(:,n)
%sum(r,1)==1, normalized probabilities
%
%NOTE: this requires that the pD.Gaussians components
%are already at least slightly different,
%otherwise all responsibilities will be identical,
%and training gets stuck....

lPz=meanlogprob(pD.MixWeight);%E[ln P[Z] ]; Eq. 10.66
[lPx E2k]=meanlogprob(pD.Gaussians,x);%E[ ln p(x | Gaussians) ]; Eq. 10.65,67
lPx=lPx+repmat(lPz,1,size(lPx,2));
lPx=lPx-max(max(lPx));%prevent underflow next
r=exp(lPx);
r=r./repmat(sum(r,1),size(r,1),1);%normalize