%==========================================================================
%==========================================================================
% Copyright (C) 2014 by Arne Leijon
% Email: leijon@kth.se

% This program is free software for academic and non-commercial
% use, to be used at user's own risk, no guarantee of performance.
% The user is permitted to copy, distribute and transmit only unaltered
% copies of this program. 
%==========================================================================
%==========================================================================
%[pD,LL,pDprior]=adapt(pDprior,x,pZ)
%method to adapt a GaussB object or array of such objects
%using training data.
%
%Input:
%pDprior=   GaussB object or array of GaussD objects
%           with prior property values
%x=         training data vectors, stored columnwise
%pZ=        component responsibilities
%pZ(i,n)=   P[ Z(n)==i | x(:,n) ],
%           where Z(n) is index of the GaussB element
%           that might have generated x(:,n)
%
%Result:
%pD=        adapted version of the GaussB object(s)
%LL=        low bound on Log-Likelihood of training data
%           in the variational approximation.
%           (Bishop, 10.2.2)
%
%Method: Bishop(2006), sec. 10.2.
%
%Arne Leijon 2007-05-25 slightly tested
%***** Svante Stadler

function [pD,LL]=adapt(pDprior,x,pZ,lPz,pDWeight,pDWinit,E2k)

minWeight=0.001;%limit to avoid underflow...???
pD=pDprior;
nObj=numel(pD);
[nD,nX]=size(x);
if nargin<3%should happen only when nObj==1
    if nObj==1
        pZ=ones(nObj,nX);%equal weight for all objects
    else
        error('GaussB array needs external component weights');
    end;
end;
LL=zeros(nObj,1);%space
for k=1:nObj%for all GaussB objects
    Nk=sum(pZ(k,:),2);%Eq. 10.51:effective number of obs. for this GaussB object
    if Nk>minWeight%we still have some weight on this component
        %UPDATE pD
        b0 = pD(k).MeanPrec;
        v0 = pD(k).WishartDF;
        W0=pD(k).WishartW;
        %                 m0 = 0;
        m0=pD(k).Mean;
        W0i = inv(pD(k).WishartW);
        xk = (1/Nk) * sum(x.*repmat(pZ(k,:),nD,1),2);
        x0 = x - repmat(xk,1,nX);
        bk = b0 + Nk;
        mk = (b0*m0 + Nk*xk)/bk;
        NkSk = (x0.*repmat(pZ(k,:),nD,1))*x0';
        Wvar = b0*Nk/bk * (xk-m0)*(xk-m0)';
        Wi = W0i + NkSk + Wvar;
        
        %set data
        pD(k).Mean      = mk;
        pD(k).WishartW  = inv(Wi);
        pD(k).MeanPrec  = b0 + Nk;
        pD(k).WishartDF = v0 + Nk;
        aaa=1;
    else%**** the new values may cause numerical underflow
        pD(k)=pDprior(k);%just copy prior
        pD(k).MeanPrec=pDprior(k).MeanPrec+minWeight;%to avoid underflow
        pD(k).WishartDF=pDprior(k).WishartDF+minWeight;%to avoid underflow
        b0 = pD(k).MeanPrec;
        v0 = pD(k).WishartDF;
        W0=pD(k).WishartW;
        m0=pD(k).Mean;
        W0i = inv(pD(k).WishartW);
        xk = (1/(eps+Nk)) * sum(x.*repmat(pZ(k,:),nD,1),2);
        x0 = x - repmat(xk,1,nX);
        bk = b0 + Nk;
        mk = (b0*m0 + Nk*xk)/bk;
        NkSk = (x0.*repmat(pZ(k,:),nD,1))*x0';
        Wvar = b0*Nk/bk * (xk-m0)*(xk-m0)';
        Wi = W0i + NkSk + Wvar;
%                 NkSk=zeros(nD,nD);
%                 bk=b0;
%                 b0=0;
%                 mk=zeros(nD,1);
%                 m0=mk;
%                 xk=mk;
%                 W0=NkSk;
        % nargout=1;
        aaa=2;
    end;
    %%%%%%%%%%%%
    if nargout>1 %calculate log-likelihood bound
        scale = max(max( pD(k).WishartW));
        Ws =  pD(k).WishartW * (1/scale);
        K = size(pD(k).WishartW,1)*log(scale);
        %         E2 = nD*log(2) + log(det(Ws))+ K + sum(psi((pD(k).WishartDF +1-(1:nD))/2));
        %         E2 = nD*log(2) + log(det( pD(k).WishartW ))+ sum(psi((pD(k).WishartDF +1-(1:nD))/2));
        E2=E2k(k);
%         E2k(k)=E2;
        E3 = -pD(k).WishartDF*trace((NkSk/(Nk+eps))*pD(k).WishartW);
        E4=-nD/bk;
        x_m2 = xk-pD(k).Mean;
        E5=-pD(k).WishartDF*(x_m2)'*pD(k).WishartW*(x_m2);
        E6=-nD*log(2*pi);
        L1k(k)=Nk*(E2+E3+E4+E5+E6);
        %%%%%%%%%%%%%%%%
        F1=nD*log(b0/(2*pi))-(nD*b0)/(bk);
        F2=-(b0*pD(k).WishartDF)*(mk-m0)'*pD(k).WishartW*(mk-m0);
        L4k_temp1(k)=F1+F2+E2;
        L4k_temp4(k)=pD(k).WishartDF*trace(inv(W0)*pD(k).WishartW);
        %%%%%%%%%%%%%%%%%%%
        
        T1=.5*E2+.5*nD*log(bk/(2*pi))-.5*nD;
%         BWk1=(det(pD(k).WishartW))^(-pD(k).WishartDF/2);
%         lnBWk1=log(BWk1);
        lnBWk1=(-pD(k).WishartDF/2)*log((det(pD(k).WishartW)));
%         BWk2=2^(pD(k).WishartDF*nD/2);
        lnBWk2=(pD(k).WishartDF*nD/2)*log(2);
%         BWk3=pi^(nD*(nD-1)/4);
        lnBWk3=(nD*(nD-1)/4)*log(pi);
        for i=1:nD
%             BWk4_temp(i)=gamma((pD(k).WishartDF+1-i)/2);
            lnBWk4_temp(i)=gammaln((pD(k).WishartDF+1-i)/2);
%             lnBWk4_temp(i)=log(BWk4_temp(i));
        end
%         BWk4=prod(BWk4_temp);
        lnBWk4=sum(lnBWk4_temp);
%         BWk5=BWk2*BWk3*BWk4;
        lnBWk5=lnBWk2+lnBWk3+lnBWk4;
%         BWk6=inv(BWk5);
        lnBWk6=-lnBWk5;
%         BWk=BWk6*BWk1;
        lnBWk=lnBWk6+lnBWk1;
%         HWk=-log(BWk)-.5*(pD(k).WishartDF-nD-1)*E2+.5*nD*pD(k).WishartDF;
        HWk2=-lnBWk-.5*(pD(k).WishartDF-nD-1)*E2+.5*nD*pD(k).WishartDF;
        Lk7(k)=T1+HWk2;
        %         Lk7(k)=.5*E2+.5*nD*log(bk)-.5*nD*log(2*pi)+log(BWk)+.5*(pD(k).WishartDF+nX)*(1-nD)-.5*(nD+1);
        %LL(k)= %*******NOT yet implemented
    else
        %         L1k=zeros(1,nObj);
        %         L4k_temp1=zeros(1,nObj);
        %
    end;
    
    %%%%%%%%%%%%
end;
if nargout>1 
    for n=1:nX
        %     L2k=0;
        for  k=1:nObj
            L2k(k)=pZ(k,n)*lPz(k);
            %         L2k=L2k+L2k_temp;
        end
        L2n(n)=sum(L2k);
    end
    L2=sum(L2n);
    L1=0.5*sum(L1k);
    
    alpha=pDWeight.Weight;
    alpha0=pDWinit;
    alpha_hat=sum(alpha);
    alpha_hat0=sum(alpha0);
    gamma_alpha=gamma(alpha);
    loggamma_alpha=gammaln(alpha);
    gamma_alpha0=gamma(alpha0);
    loggamma_alpha0=gammaln(alpha0);
    % C_alpha=gamma(alpha_hat)/prod(gamma_alpha);
    logC_alpha=gammaln(alpha_hat)-sum(loggamma_alpha);
    % C_alpha0=gamma(alpha_hat0)/prod(gamma_alpha0);
    logC_alpha0=gammaln(alpha_hat0)-sum(loggamma_alpha0);
    
    L3=logC_alpha0+((alpha0(1)-1)*sum(lPz));
    BW1=(det(W0))^(-v0/2);
    BW2=2^(v0*nD/2);
    BW3=pi^(nD*(nD-1)/4);
    for i=1:nD
        BW4_temp(i)=gamma((v0+1-i)/2);
    end
    BW4=prod(BW4_temp);
    BW5=BW2*BW3*BW4;
    %----------- Jalal
    %     BW6=inv(BW5);
    BW6=BW5^-1;
    %---------------------
    BW=BW1*BW6;
    L4_temp2=nObj*log(BW);
    L4_temp3=.5*(v0-nD-1)*sum(E2k);
    L4_temp1=0.5*sum(L4k_temp1);
    L4_temp4=-0.5*sum(L4k_temp4);
    L4=L4_temp1+L4_temp2+L4_temp3+L4_temp4;
    for n=1:nX
        for  k=1:nObj
            LpZk(k)=pZ(k,n)*log(pZ(k,n)+eps);
        end
        LpZn(n)=sum(LpZk);
    end
    L5=sum(LpZn);
    
    for k=1:nObj
        L6_temp(k)=(alpha(k)-1)*lPz(k);
    end
    L6=sum(L6_temp)+logC_alpha;
    
    L7=sum(Lk7);
    
    LL=L1+L2+L3+L4-L5-L6-L7;
%     LL=L1;
else
%     LL=LL2;
end

% LL=L;%across all Gaussian components