%logP=logprob(pD,x) gives posterior log(probability densities) for given vectors
%assumed to be drawn from a given GaussB object.
%Probability is marginalized over the posterior distribution of random
%parameters.
%
%Input:
%pD=    GaussB object or array of such objects
%x=     matrix with given vectors, stored columnwise
%
%Result:
%logP=  log(probability densities for x)
%       if pD is one-dimensional vector,
%           size(logP)== [length(pD),size(x,2)].
%       if pD is multidim array,
%           size(logP)== [size(pD),size(x,2)].
%exp(logP)= true probability densities for x
%
%The log representation is useful because the probability densities may be
%extremely small for random vectors with many elements.
%
%Method:
%After Bayesian training, the predictive density, given only training data,
%is a (multidimensional) Student density.
%Ref: Bishop(2006), 10.2 + App B
%  
%Arne Leijon 2007-05-25 slightly tested
%           2008-01-22 numerically improved

function logP = logprob(pD,x)
pDsize = size(pD);%size of GaussB array
nObj = numel(pD);%number of GaussB objects in array
nx = size(x,2);%number of observed vectors
logP=zeros(nObj,nx);%space
for i = 1 : nObj%for all GaussB objects in pD
    D = length(pD(i).Mean);%GaussB random vector size
    if D == size(x,1)%observed vector size OK
        v = pD(i).WishartDF + 1 - D;
        lambda = v * pD(i).MeanPrec / (1 + pD(i).MeanPrec) * pD(i).WishartW;
        logdetlambda=sum(log(eig(lambda)));%prevent overflow
        logP(i,:) = gammaln(v/2 + D/2) - gammaln(v/2) + 1/2 * logdetlambda - D/2 * log(v * pi);
        dx=x - repmat(pD(i).Mean,1,nx);
        Delta2=sum(dx.*(lambda * dx),1);
        logP(i,:) = logP(i,:) - (v + D)/2 * log(1 + Delta2 / v);
    else
        warning('Incompatible data size');
        logP(i,:) = repmat(-Inf,1,nx);%zero probability
    end;
end;
if nObj > 1
    logP = squeeze(reshape(logP,[pDsize,nx]));%restore array format
end;%not needed if pD is a vector