function [d, d_raw] = MIVB(x_cs, y_ps, fs_signal, varargin)
%==========================================================================
%==========================================================================
% MIVB is Copyright (C) 2014 by Jalal Taghia and Rainer Martin
% at Institute of Communication Acoustics (IKA),
% Ruhr-Universitt Bochum.

% This program is free software for academic and non-commercial
% use, to be used at user's own risk, no guarantee of performance.
% The user is permitted to copy, distribute and transmit only unaltered
% copies of this program.
%==========================================================================
%==========================================================================
% MIVB instrumental intelligibility measure for MATLAB
% Version 1.0, October 14, 2014
% Implemented by JALAL TAGHIA
% Email: jalal.taghia@rub.de

% This program implements the instrumental intelligibility measure MIVB.
% First, input signals are transformed into 15 subbands by using a 1/3 octave
% band filter bank and then mutual information between the amplitude envelopes
% of the reference and the test signals is estimated per subband by using a
% variational Bayesian inference based GMM. For more details on the method, see [1].

% The intelligibility score (in unit of nats) is derived by taking the average
% of all values over subbands.

% In the normalization stage, the intelligibility score between the test signal
% and the reference signal is normalized with the intelligibility score
% between the reference signal and itself which has the maximum information.
% The normalized intelligibility score is presented in percent (%).

% REFERENCE:
% [1] Jalal Taghia and Rainer Martin, "Objective intelligibility measures based
% on mutual information for speech subjected to speech enhancement processing",
% IEEE Transactions on Audio, Speech, and Language Processing, vol. 22,
% no. 1, pp. 6-16, January 2014.
%==========================================================================
% MAIN INPUTS:
%             x_cs          a vector of clean speech (reference) signal or
%                           a matrix of clean speech (reference) signals.
%                           x_cs is always considered as the reference for
%                           measuring the intelligibility of the test signals.
%                           If x_cs is chosen a matrix of reference
%                           signals, then the size the matrix x_cs must be the
%                           same as the size of the matrix of test signals
%                           y_ps.
%                           If x_cs is a vector containing one reference
%                           signal then all the test signals in the matrix
%                           y_ps are evaluated with respect to the single
%                           reference signal in x_cs.

%             y_ps          a vector of test signal or a matrix of test signals
%                           (e.g., noisy and processed speech signals).

%             fs_signal     sampling frequency of input signals in Hz.


% OPTIONAL INPUTS:

%            'sr'           silence removal, can be set 'on' or 'off'. By
%                           default, it is 'on' i.e., silence removal is
%                           performed. You can set it to 'off' if there is almost no
%                           speech pause in the reference signal(s).

%           'com_ini'       initial number of mixture components for the
%                           variational Bayesian based GMM. It is set by
%                           default according to the length of input signals.

%           'num_iter'      number of iterations for the parameter
%                           estimation of GMM in a variational Bayes framework.
%                           It is set by default according to the length
%                           of input signals.

%==========================================================================
% MAIN OUTPUT
%            d              normalized value of instrumental intelligibility
%                           score in percent (%) for test signals.
%                           This is the main output of the function. It takes
%                           values between 0 and 100.

%                           The normalized instrumental score d is computed
%                           by dividing the intelligibility score between
%                           the test signal and the reference signal with
%                           the intelligibility score obtained between the
%                           reference signal and itself.

%                           Hence, it shows the percentage of information
%                           which the test signal contains about the reference
%                           signal.
%                           It has been shown in [1] that this measure has
%                           a high performance in predicting the intelligibility
%                           of speech subjected to speech enhancement processing.
%                           Its value should not be misinterpreted with the percent
%                           value of the word correct score (WCS) which is defined
%                           in a subjective listening test.
%                           For the latter, a map function is required to find the
%                           equivalent values for word correct scores.
%                           This program does not provide the map function.
%                           The map function should be derived based on the
%                           existing corpus and the noise conditions
%                           in the experiments.


% ADDITIONAL OUTPUT

%            d_raw          the value of instrumental intelligibility score
%                           for test signal in nats before the normalization
%                           takes place. The unit of d_raw is in nats.
%==========================================================================
% Usage Examples
%                           [d] = MIVB(x_cs, y_ps, fs_signal);
%                           The simplest use of function (the default setting).
%                           In this case, the silence removal is 'on'.
%                           The initial number of mixture components com_ini and
%                           the number of iterations num_iter are set according
%                           to the input signal's length.


%                           [d] = MIVB(x_cs, y_ps, fs_signal,'sr','off');
%                           In this way, the silence removal is turned 'off'.

%                           [d] = MIVB(x_cs, y_ps, fs_signal,'com_ini',30, 'num_iter', 1000);
%                           In this way, you can set an arbitrary value for
%                           the initial number of mixture components com_ini
%                           and the number of iterations num_iter.

%                           [d,d_raw] = MIVB(x_cs, y_ps, fs_signal,'sr','on','com_ini',30, 'num_iter', 1000);
%                           The usage of function in a complete form.
%==========================================================================
%==========================================================================
% check the length of input signals
if length(x_cs) ~= length(y_ps)
    error('Inpunt signals should have the same length!');
end

if size(y_ps,1) <= size(y_ps,2)
    y_ps = y_ps';                        % test signals are placed in columns
end

if size(x_cs,1) <= size(x_cs,2)
    x_cs = x_cs';                        % reference signals are placed in columns
end

if size(x_cs,2)==1
    v_flag = 'on';                 % there is a single reference signal in the input
    x_cs = repmat(x_cs(:,1),1,size(y_ps,2));
elseif (size(x_cs,2) < size(y_ps,2)) || (size(x_cs,2) > size(y_ps,2))
    v_flag = 'off';
    error(['>> The reference input x_cs should be either a vector or a',...
        ' matrix with the same size as the matrix of test signals y_ps.']);
else
    v_flag = 'off';
end
%--------------------------------------------------------------------------
% Default setting
if nargin<3
    error('>> Provide at least three inputs!');
end

sr = 'on'; % silence removal flag. By default, silence removal is
%           performed, you can set it to 'off' if there is almost no speech pause in
%           reference signal.
com_ini_flag = 'off';  % by default com_ini is set according to the signal length
num_iter_flag = 'off'; % by default num_iter is set according to the signal length
%--------------------------------------------------------------------------
% check input arguments
if nargin == 3
    sr = 'on';
    display(['>> By default, the silence removal is performed. You can set it to "off"' ...
        ' if there is almost no speech pause in referece signal!']);
    string = sprintf(['>> By default, the initial number of mixture components com_ini and the number of \n' ...
        'iterations num_iter for GMM are computed according to the length of input signals.']);
    disp (string);
else
    if (rem(length(varargin),2)==1)
        error('>> Optional parameters should always go by pairs!');
    end
    
    for i=1:2:(length(varargin)-1)
        if ~ischar (varargin{i}),
            error (['>> Unknown type of optional parameter name (parameter' ...
                ' names must be strings).']);
        end
        switch lower (varargin{i})
            case 'sr'
                sr = lower (varargin{i+1});
                if strcmp(sr, 'on')
                    display('>> Silence removal is performed ...');
                elseif strcmp(sr, 'off')
                    display('>> Silence removal is not performed ...');
                else
                    error('>> Decide on silence removal by choosing "on" or "off"!');
                end
            case 'com_ini'
                com_ini = round(varargin{i+1});
                com_ini_flag = 'on';
            case 'num_iter'
                num_iter = round(varargin{i+1});
                num_iter_flag = 'on';
            otherwise
                error ('>> Provide a valid optional input argument!');
        end
    end
end
%--------------------------------------------------------------------------
% Check the length of input signal
if strcmp(sr,'off')
    if (length(x_cs)/fs_signal)<1
        error('Input signal(s) is (are) too short (i.e., less than 1s)!');
    elseif ( (length(x_cs)/fs_signal)>=1 ) && ( (length(x_cs)/fs_signal)<2 )
        warning('>> Input signal(s) is (are) too short (less than 2 s), causing inaccurate result!')
    else
    end
    sig_leng = length(x_cs);
    
elseif strcmp(sr,'on')
    x_np_len = zeros(1,size(x_cs,2));
    for kk = 1:size(x_cs,2)
        x_np = sr_func(x_cs(:,kk),x_cs(:,kk) , 40, (0.050*fs_signal));
        %              for silence removal: dynamic range 40 dB and frame length 50 ms
        x_np_len(kk)=length(x_np);
        if (x_np_len(kk)/fs_signal)<1
            error(['>> The speech-active part of the ',num2str(kk),...
                ' th reference signal is too short (less than 1 s)!']);
        elseif ( (x_np_len(kk)/fs_signal)<2 ) && ( (x_np_len(kk)/fs_signal)>=1 )
            warning(['>> The speech-active part of the ',num2str(kk),...
                ' th reference signal is too short (less than 2 s), causing inaccurate result!']);
        else
        end
    end
    sig_leng = mean(x_np_len);
else
end
%--------------------------------------------------------------------------
% check the parameters com_ini and num_iter
if strcmp(com_ini_flag,'on') && strcmp(num_iter_flag,'off')
    num_iter = com_ini * 50;
    display(['com_ini =',num2str(com_ini),'.']);
    display(['num_iter =',num2str(num_iter),'.']);
    display('You can change the number of iterations by setting a value for num_iter!');
    
elseif strcmp(com_ini_flag,'on') && strcmp(num_iter_flag,'on')
    display(['num_iter =',num2str(num_iter),'.']);
    display(['com_ini =',num2str(com_ini),'.']);
    
elseif strcmp(com_ini_flag,'off')
    if (sig_leng/fs_signal)<=30 % for signals shorter than 30 s
        com_ini = 20;
        if strcmp(num_iter_flag,'off')
            num_iter = 1000;
        end
    elseif ((sig_leng/fs_signal)<=60) && ((sig_leng/fs_signal)>30) % for signals shorter than 60 s
        com_ini = 40;
        if strcmp(num_iter_flag,'off')
            num_iter = 2000;
        end
    elseif (sig_leng/fs_signal)>60 % for signals longer than 60s
        com_ini = ((sig_leng/fs_signal)*40)/60;
        com_ini = round(com_ini );
        if strcmp(num_iter_flag,'off')
            num_iter = com_ini*50;
        end
    else
    end
else
    error('');
end
display('>> Program is running (the running time depending on the length of input signals may take a few minutes)...');
%--------------------------------------------------------------------------
% computing instrumental scores
N_test = size(y_ps,2);
d_vec = zeros(1,N_test);
for test =1:N_test
    x = x_cs(:,test);
    y = y_ps(:,test);
    if test ==1
        [d_vec(1,test), com_ini, num_iter] ...
            = intellscoreVB(x,y, sr, com_ini,com_ini_flag, num_iter,num_iter_flag , fs_signal);
        % directly the estimated com_ini and num_iter is used for the other signals
        com_ini_flag = 'on';
        num_iter_flag = 'on';
    else
        d_vec(1,test) ...
            = intellscoreVB(x,y, sr, com_ini,com_ini_flag, num_iter,num_iter_flag , fs_signal);
    end
end
% normalization of intelligibility scores
if strcmp(v_flag,'on')
    Nfactor ...
        = intellscoreVB(x_cs(:,1),x_cs(:,1), sr, com_ini,com_ini_flag, num_iter,num_iter_flag, fs_signal );
elseif strcmp(v_flag,'off')
    [AA,uniq,ccc] = unique(x_cs','rows'); % finds signals which are unique!
    N_uniq = length(uniq); % number of unique signals
    Nfactor_vec = zeros(1,N_uniq);
    for ii=1:N_uniq
        Nfactor_vec(1,ii) ...
            = intellscoreVB( x_cs(:,uniq(ii)) , x_cs(:,uniq(ii)) , sr, com_ini, com_ini_flag, num_iter, num_iter_flag, fs_signal );
    end
    Nfactor = max(Nfactor_vec);
else
end
d = 100 * (d_vec / Nfactor); % normalized instrumental intelligibility score(s)
%                             in percentage for test signal(s).
d_raw = d_vec;              % instrumental intelligibility score(s) in nats
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%    subfunctions    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% intellscoreVB
function [MI, com_ini, num_iter] = intellscoreVB(x, y, sr, com_ini, com_ini_flag, num_iter, num_iter_flag, fs_signal)
% initialization for DFT analysis and the one-third octave filter bank
fs          = 10000;                            % sample rate of proposed intelligibility measure
frame_len   	= 256;                              % window support or window length
% here frame update length is 50%
K           = 512;                              % FFT size
J           = 15;                               % Number of 1/3 octave bands
mn          = 150;                              % Center frequency of first 1/3 octave band in Hz.
H           = onethird_filterbank(fs, K, J, mn);           % Get 1/3 octave band matrix
%--------------------------------------------------------------------------
% resample signals if other samplerate is used than fs
if fs_signal ~= fs
    x	= resample(x, fs, fs_signal);
    y 	= resample(y, fs, fs_signal);
end
%----------------------------------------------------------------------
% remove silent frames
if strcmp(sr, 'on')
    dyn_range=40; % in dB (human speech is normally perceived over a range of about 40 dB)
    N_frame    	= 500; % 50 ms frame length
    %     %     figure
    %     %     subplot(2,1,1),plot(x)
    [x, y] = sr_func(x, y , dyn_range, N_frame);
    %     %     subplot(2,1,2),plot(x,'r')
    %     %     soundsc(x,fs)
    %     %     pause
end
%----------------------------------------------------------------------
% apply 1/3 octave band TF-decomposition
x_hat     	= Analysis_DFT(x, frame_len, frame_len/2, K); 	% apply short-time DFT to clean speech
y_hat     	= Analysis_DFT(y, frame_len, frame_len/2, K); 	% apply short-time DFT to processed speech

x_hat       = x_hat(:, 1:(K/2+1)).';         	% take clean single-sided spectrum
y_hat       = y_hat(:, 1:(K/2+1)).';        	% take processed single-sided spectrum
%--------------------------------------------------------------------------
X           = zeros(J, size(x_hat, 2));
Y           = zeros(J, size(y_hat, 2));

for i = 1:size(x_hat, 2)
    X(:, i)	= sqrt(H*abs(x_hat(:, i)).^2);      % apply 1/3 octave bands
    Y(:, i)	= sqrt(H*abs(y_hat(:, i)).^2);
end

% MI estimation by Bayesian approach
%00000000000000000000000000000000000000000000000000000000000000000000000000
T_select = length(X);
MI = zeros(1, J);
com = zeros(1, J);
for sub=1:J
    if strcmp(com_ini_flag, 'off')
        if sub==1
            display('>> The number of mixture components is estimated ...');
        end
        [MI(1, sub), com(1, sub)] = MIpara_Bayesian(X(sub, :), Y(sub, :), com_ini, num_iter, T_select);
        if (sub==1)
            com_ini = com(1, sub)+ 5;
            if strcmp(num_iter_flag, 'off')
                num_iter = com_ini*50;
            end
            display(['com_ini =',num2str(com_ini)]);
            display(['num_iter =',num2str(num_iter)]);
            [MI(1, sub), com(1, sub)] = MIpara_Bayesian(X(sub, :), Y(sub, :), com_ini, num_iter, T_select);
        end
    elseif strcmp(com_ini_flag, 'on') && strcmp(num_iter_flag, 'on')
        [MI(1, sub), com(1, sub)] = MIpara_Bayesian(X(sub, :), Y(sub, :), com_ini, num_iter, T_select);
    else
    end
end
MI=mean(max(MI,0)); % Here we put a limit for minimum MI to zero for all bands
end
%==========================================================================
%==========================================================================
%% onethird_filterbank
function  FBmat = onethird_filterbank(fs, DFTlen, NB, fcf)
%   outputs:
%       FBmat:          matrix of 1/3 octave bands

%   inputs:
%       fs:         sampling frequency (Hz)
%       DFTlen:      DFT length
%       NB:         number of subbands
%       fcf:         center frequency of first 1/3 octave band in Hz
freq_vec = linspace(0, fs, DFTlen+1);
freq_vec = freq_vec(1:(DFTlen/2+1));
NumFbin = length(freq_vec);
k = 0:(NB-1);
cf_k  = 2.^(k/3)*fcf;
cf_ku  = 2.^((k+1)/3)*fcf;
cf_kd  = 2.^((k-1)/3)*fcf;
fd = sqrt(cf_k.*cf_kd);
fu = sqrt(cf_k.*cf_ku);
FBmat = zeros(NB, NumFbin);

for i = 1:NB
    [aaa, nf_ind]  = min((freq_vec-fd(i)).^2);
    fd(i)  = freq_vec(nf_ind);
    fd_ii  = nf_ind;
    %--------------------------%
    [aaa, nf_ind] = min((freq_vec-fu(i)).^2);
    fu(i) = freq_vec(nf_ind);
    fu_ii = nf_ind;
    FBmat(i,fd_ii:(fu_ii-1))= 1;
end

NF_active  = sum(FBmat, 2);
% finding actual number of subbands
NB_ac = find( ( NF_active(2:end) >= NF_active(1:(end-1)) ) & ( NF_active(2:end)~=0 ) ~=0, 1, 'last' )+1;
FBmat   = FBmat(1:NB_ac, :);
end
%==========================================================================
%==========================================================================
%% Analysis_DFT
function X_DFT = Analysis_DFT(x, N, K, DFTlen)
% Inputs:
%           x:      input signal (a vector)
%           N:      frame length
%           K:      overlap between frame_ind
% Outputs:
%           X_DFT:  DFT coefficients of the imput signal where its rows and
%                   columns indicate the DFT bin and time-frame indices
window  = hanning(N);
x = x(:);
frame_ind = 1:K:(length(x)-N);
Num_frame = length(frame_ind);
X_DFT     = zeros(Num_frame, DFTlen);
for i = 1:Num_frame
    sample_ind = frame_ind(i):(frame_ind(i)+N-1);
    win_x = x(sample_ind).*window;
    X_DFT(i, :) = fft(win_x, DFTlen);
end
end
%==========================================================================
%==========================================================================
%% sr_func
function [x_a ,y_a] = sr_func(x, y, dyn_range, N)
%   x:      reference input signal (a vector)
%   y:      the second input signal which in the reconstruction follows the
%           reference signal x.
%   N:      frame-length (number of samples samples)
%   dyn_range: dynamic range in dB

N= floor(N);
x = x(:); y  = y(:);
lenx = floor(length(x)/N)*N;
xmat = reshape(x(1:lenx),N,lenx/N);
xx = x(lenx+1:end);
Exx = xx.^2;
Exxmax = mean(Exx);
Exmat = xmat.^2;
Exmax = max([mean(Exmat),Exxmax]);
Ex = .5*db(x.^2);
Exmax = .5 * db(Exmax);
frame_get = (Ex>(Exmax - dyn_range));
x_a = x(frame_get);
y_a = y(frame_get);
end
%==========================================================================
%==========================================================================
%% MIpara_Bayesian
% this function is for multi components
function [MI, com] = MIpara_Bayesian(x, y, com_ini, num_iter, M_sample)

if size(x,1)>size(x,2)
    x=x';
end
if size(y,1)>size(y,2)
    y=y';
end
Dx=size(x,1);
Dy=size(y,1);
[px_hat_sample, py_hat_sample, pz_hat_sample, com] = ...
    GMM_multiD_Bayesian(x, y, com_ini, num_iter, M_sample);

% Estimating MI
px_hat_sample(px_hat_sample==0)=eps;
py_hat_sample(py_hat_sample==0)=eps;
pz_hat_sample(pz_hat_sample==0)=eps;
MI = mean (log ((pz_hat_sample)./(px_hat_sample.*py_hat_sample)));
end
%==========================================================================
%==========================================================================
%% GMM_multiD_Bayesian
function [px_hat_sample_sum, py_hat_sample_sum, pz_hat_sample_sum, com] = ...
    GMM_multiD_Bayesian(x, y, com_ini, num_iter, M_sample)
z=[x;y];
%===============Deriving the model ========================================
% GMM based on variational Bayesian (from toolbox provided by Arne Leijon)
g=init(GMMB(com_ini), z);
[g,LL,r]=adapt(g, z, num_iter, Inf);
% =========================================================================
mix = floor(double(g.MixWeight.Weight));

EffecCom_ind = find(mix>0);
com = length(EffecCom_ind);
[D, T]=size(z);
Dx=size(x,1);
Dy=size(y,1);
model_mu = zeros(D, com_ini);
model_Sigma = zeros(D,D,com_ini);
for ind = 1:com_ini
    model_mu(:,ind) = g.Gaussians(ind, 1).Mean;
    if g.Gaussians(ind, 1).WishartDF < D+1
        model_Sigma(:,:,ind) = (g.Gaussians(ind, 1).WishartDF) .* ...
            (g.Gaussians(ind, 1).WishartW); % mean of Wishart dist
    else
        model_Sigma(:,:,ind) = ((g.Gaussians(ind, 1).WishartDF)-D-1) .* ...
            (g.Gaussians(ind, 1).WishartW); % Mode of Wishart dist
    end
    model_Sigma(:,:,ind) =model_Sigma(:,:,ind)^-1; % Before taking the inverse it is the precesion
    % not the covariance.
end
model_weight = g.MixWeight.Mean;
% To take smaple ------------------------------
% joint distribution x and y
[pz_hat_sample_sum] = GMM_form_multiCom_special (model_weight, model_mu, model_Sigma, z(:,1:M_sample));
% marginal distribution x
[px_hat_sample_sum] = GMM_form_multiCom_special (model_weight, model_mu(1:Dx,:), model_Sigma(1:Dx, 1:Dx, :), x(:,1:M_sample));
% marginal distribution y
[py_hat_sample_sum] = GMM_form_multiCom_special (model_weight, model_mu(Dx+1:Dy+Dx,:), model_Sigma(Dx+1:Dy+Dx, Dx+1:Dy+Dx, :), y(:,1:M_sample));
end
%==========================================================================
%==========================================================================
%% GMM_form_multiCom_special
function [pz_hat] = GMM_form_multiCom_special (modelweight, modelmu, Covmat, L)
T=length(L);
[D, com]=size(modelmu);
pz_hat=zeros(1,T);
pz_hat_ith=zeros(com,T);
for t=1:T
    for i=1:com
        modelSigma=Covmat(:,:,i);
        term = 1/(((2*pi)^(D/2))*((det(modelSigma))^.5));
        pz_hat_ith(i,t)=modelweight(i) * term * exp(-.5 * (L(:,t)-modelmu(:,i))'/ modelSigma * (L(:,t)-modelmu(:,i)));
    end
end
pz_hat=sum(pz_hat_ith,1);
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% THE END   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

